import os
from MDAnalysis import Universe
import numpy as np
try:
    # for openmm version >= 7.6
    from openmm.app import pdbxfile, topology, element, pdbfile
except:
    # for openmm version <= 7.5
    from simtk.openmm.app import pdbxfile, topology, element, pdbfile

_DNA_RES = 'BP'
_DNA_S1_TYPE = 'I'
_DNA_S2_TYPE = 'J' # J atoms in reversed order to pair with I
_PRO_RES = 'HIT'
_PRO_TYPE = 'C'
_VIRTUAL_TYPE = 'V'

def coarse_grain_atompdb_2_wechromcif(pdb_filename, out_dir = os.getcwd(), out_file_prefix = 'coarse_grain', out_pdb = False):
    """Coarse grain an atomistic DNA pdb file into a wechrom cif file.
    ***YOU MAY NEED TO MODIFY THIS FUNCTION DEPENDING ON YOUR PDB FILE***

    Args:
        pdb_filename (str or file): pdb file
        out_dir (str, optional): output directory. Defaults to os.getcwd().
        out_file_prefix (str, optional): output file prefix. Defaults to 'coarse_grain'.
        out_pdb (bool, optional): whether to generate a corase grained pdb file. Defaults to False.
    """
    u = Universe(pdb_filename)

    # read DNA and corase grain 
    # ***You may need to modify this section depending on your pdb file***
    nucleic_atoms = u.select_atoms("nucleic")
    I = nucleic_atoms.segments[0].atoms
    J = nucleic_atoms.segments[1].atoms
    I_pos = np.array([res.atoms.center_of_mass() for res in I.residues]) 
    J_pos = np.array([res.atoms.center_of_mass() for res in reversed(J.residues)]) # reverse order
    
    # convert to reduced units
    I_pos /= 0.241
    J_pos /= 0.241

    # sanity check
    n_bp = I_pos.shape[0]
    assert n_bp == J_pos.shape[0], "I and J strands should have same number of beads"  

    # build wechrom topology 
    topo = topology.Topology()
    ele_DNA = element.Element.getBySymbol("N")

    # add DNA atoms to chain 0
    # each DNA residue has name "BP"
    # each DNA atom has name "I"/"J", element "N"
    chain = topo.addChain(0) 
    for i in range(n_bp):
        residue = topo.addResidue(name = "BP", chain = chain) 
        atom = topo.addAtom(name = "I", element=ele_DNA, residue=residue)
        atom = topo.addAtom(name = "J", element=ele_DNA, residue=residue)

    positions = [None] * (2*n_bp)
    positions[::2] = I_pos
    positions[1::2] = J_pos
    
    # set unit cell dimensions
    dim = np.absolute(np.amax(positions, axis = 0) - np.amin(positions, axis=0))
    topo.setUnitCellDimensions(np.round(dim*2))

    # write corase grained files
    with open(os.join(out_dir, out_file_prefix+".cif"), "w") as fo:
        pdbxfile.PDBxFile.writeFile(topo, positions, fo)
    if out_pdb:
        with open(os.join(out_dir, out_file_prefix+".pdb"), "w") as fo:
            pdbfile.PDBFile.writeFile(topo, positions, fo)

def num_intra_memory_bonds(n_bp, memory_range, n_template):
    """Calculate the expected number of bonds for intra-strand associative memory force

    Args:
        n_bp (int): number of basepairs in the system
        memory_range (int): size of the memory
        n_template (int): number of templates used as memories

    Returns:
        int: number of expected intra-strand bonds
    """
    n_bonds = 0
    for i in range(1, memory_range):
        n_bonds += (n_bp - i) * (memory_range - i)
    n_bonds *= 2 * n_template
    return n_bonds
    
def num_inter_memory_bonds(n_bp, memory_range, n_template):
    """Calculate the expected number of bonds for inter-strand associative memory force

    Args:
        n_bp (int): number of basepairs in the system
        memory_range (int): size of the memory
        n_template (int): number of templates used as memories

    Returns:
        int: number of expected inter-strand bonds
    """
    n_bonds = 0
    for i in range(1, memory_range):
        n_bonds += (n_bp - i) * (memory_range - i)
    n_bonds *= 2 * n_template
    n_bonds += n_bp * memory_range * n_template
    return n_bonds


import os
import pkg_resources
import pandas as pd
import numpy as np
from collections import defaultdict
from MDAnalysis import Universe
try:
    # for openmm version >= 7.6
    from openmm.app import PDBxFile
except:
    # for openmm version <= 7.5
    from simtk.openmm.app import PDBxFile
from .utils import _DNA_S1_TYPE, _DNA_S2_TYPE, _PRO_TYPE

_DNA_TYPES = [_DNA_S1_TYPE, _DNA_S2_TYPE]

def read_memory_template(template_list_dir = 'DEFAULT', template_list_filename = 'data/naked_database_list'):
    """Read the list of associative memory template filenames

    Args:
        template_list_dir (str, optional): directory of the template list file. Defaults to 'DEFAULT'.
        template_list (str, optional): name of the template list file. Defaults to 'data/naked_database_list'.

    Raises:
        Exception: custom template list is not supported right now

    Returns:
        pandas.DataFrame: fist column 'filename' and second column 'n_bp'
    """
    if template_list_dir == 'DEFAULT':
        template_location = pkg_resources.resource_filename(__name__, template_list_filename)
    else:
        raise Exception("Custom memory template not implemented yet")
    memory_list = pd.read_csv(template_location, sep = '\s+')
    return memory_list

def generate_naked_memory_files(memory_dir = 'DEFAULT', memory_filename = 'naked.mem', memory_range = 4):
    """generate the file to store associative memory fragments

    Args:
        memory_dir (str, optional): Defaults to current workin directory.
        memory_filename (str, optional): Defaults to 'naked.mem'.
        memory_range (int, optional): Defaults to 4.
    """
    print("Preparing the associative memory files......", end = ' ')
    if memory_dir == 'DEFAULT':
        memory_dir = os.getcwd()
    # Save the memory fragments in desired file
    memory_location = os.path.join(memory_dir, memory_filename)
    with open(memory_location, 'w') as mo:
        mo.write('[Target]\nnaked DNA\n[Memories]\n')
        mo.write('filename template_res memory_range\n')

    # Read the template dataframe
    memory_list = read_memory_template(template_list_filename = 'data/naked_database_list')

    # Generate memory fragments from the memory templates
    with open(memory_location, 'a') as mo:
        for _, row in memory_list.iterrows():
            fragment = (row['n_bp'])//2
            if fragment + memory_range < row['n_bp']:
                mo.write(f"{row['filename']} {fragment} {memory_range}\n")

    print("done")

def get_naked_memory_template_bonds(memory_list_dir = 'DEFAULT', memory_list_filename = 'naked.mem', database_location = 'data/memory_template/'):
    """Extract the positions of template associative memory fragments

    Args:
        memory_list_dir (str, optional): directory of the memory list. Defaults to current workin directory.
        memory_list_filename (str, optional): filename of the memory list. Defaults to 'naked.mem'.
        database_location (str, optional): location of the database. Defaults to 'data/memory_template/'.

    Returns:
        dict, 10 x 2 x memory_range x 3: positions of templates
            layer 1(10): template, keys 0-9
            layer 2(2): dna type (strand), keys 'I', 'J'
            layer 3(np.array, memory_range x 3): positions of template residues in the range of memory_range
    """
    if memory_list_dir == 'DEFAULT':
        memory_list_dir = os.getcwd()
    # read the associative memory list files
    memory_list_location = os.path.join(memory_list_dir, memory_list_filename)
    memory_list = pd.read_csv(memory_list_location, skiprows=4, sep='\s+', names=['filename', 'template_res', 'memory_range'])
    
    template_positions = {}
    # read each associative memory fragment
    for index, row in memory_list.iterrows():
        fragment_filename = pkg_resources.resource_filename(__name__, database_location + row['filename'])

        # read the memory file
        template_positions[index] = {}
        frag = Universe(PDBxFile(fragment_filename))

        # Read the bps within the memmory_length for each strand
        for dna_type in _DNA_TYPES:
            template_query = f"resid {row['template_res']}:{row['template_res'] + row['memory_range'] - 1} and name {dna_type}"     
            template_atoms = frag.select_atoms(template_query)
            # get the positions
            template_positions[index][dna_type] = template_atoms.positions / 10 # MDanalysis default units Angstrom, convert to openmm default units nm

    return template_positions

def generate_nucleo_memory_files(memory_dir = 'DEFAULT', memory_filename = 'nucleo.mem', database_location = 'data/memory_template/', neighbor_radius = 100.0):
    """generate the file to store associative memory fragments for nucleosomes

    Args:
        memory_dir (str, optional): Defaults to current workin directory.
        memory_filename (str, optional): Defaults to 'nucleo.mem'.
        database_location (str, optional): Defaults to 'data/memory_template/'.
        neighbor_radius (float, optional): the radius to select neighbor contacts. in Angstrom for MDAnalysis. Defaults to 100.0
    """
    print("Preparing the nucleosome associative memory files......", end = ' ')
    if memory_dir == 'DEFAULT':
        memory_dir = os.getcwd()
    
    # Read the template dataframe
    memory_list = read_memory_template(template_list_filename = 'data/nucleo_database_list')

    # Save the memory fragments in desired file
    memory_location = os.path.join(memory_dir, memory_filename)
    with open(memory_location, 'w') as mo:
        mo.write('[Target]\nnucleosome\n[Memories]\n')
        mo.write('filename nucleo_center type1 res1 type2 res2_list\n')

    file_query = defaultdict(dict)
    file_center = {}
    # Generate memory fragments from the memory templates
    with open(memory_location, 'a') as mo:
        for _, row in memory_list.iterrows():
            filename = row['filename'] 
            nucleo_center = row['nucleo_center']
            contacts = [int(x) for x in row['contacts'][1:-1].split(",")]
            dna_type = row['type']

            # generate center memory (between hitone and contact DNA)
            type1 = _PRO_TYPE
            res1 = row['histone_res']
            type2 = dna_type
            res2_list = contacts            
            mo.write(f"{filename} {nucleo_center} {type1} {res1} {type2} [" + ",".join([str(int(x)) for x in res2_list]) + "]\n")

            # prepare for neighbor memory
            file_center[filename] = nucleo_center

            # prepare query to select contact atoms from the file
            file_query[filename][dna_type] = f"name {dna_type} and ("
            for res_id in contacts:
                file_query[filename][dna_type] += f"resid {res_id + 1} or " #openmm index starts from 0, mda from 1
            file_query[filename][dna_type] = file_query[filename][dna_type][:-4] + ")"
    
    for filename in file_query:
        # go through memory files one by one
        fragment_filename = pkg_resources.resource_filename(__name__, database_location + filename)
        frag = Universe(PDBxFile(fragment_filename))

        # select atoms for the two strands
        dna_atoms = {}
        for dna_type in _DNA_TYPES:
            dna_atoms[dna_type] = frag.select_atoms(file_query[filename][dna_type])

        for res1_type in _DNA_TYPES:
            for res2_type in _DNA_TYPES:
                res1_res2_set = defaultdict(set)
                res1_atoms = dna_atoms[res1_type]
                res2_atoms= dna_atoms[res2_type]
                for atom1 in res1_atoms:
                    for atom2 in res2_atoms:
                        res1_id = atom1.resid-1 #openmm index from 0, mdanalysis index from 1
                        res2_id = atom2.resid-1
                        if res2_id < res1_id:
                            continue
                        if res2_id == res1_id and res1_type == res2_type:
                            continue
                        pos1 = atom1.position
                        pos2 = atom2.position
                        rm = np.linalg.norm(pos1 - pos2)
                
                        if rm < neighbor_radius:
                            res1_res2_set[res1_id].add(res2_id)
                for res1_id in res1_res2_set:
                    res1_res2_set[res1_id] = sorted(res1_res2_set[res1_id])
                    res2_list = "["+",".join([str(int(x)) for x in res1_res2_set[res1_id]])+"]"
                    with open(memory_location, 'a') as mo:
                        mo.write(f"{filename} {file_center[filename]} {res1_type} {res1_id} {res2_type} {res2_list}\n")
    print("done")

def get_nucleo_memory_template_bonds(memory_list_dir = 'DEFAULT', memory_list_filename = 'nucleo.mem', database_location = 'data/memory_template/'):
    """Extract the positions of template associative memory fragments

    Args:
        memory_list_dir (str, optional): directory of the memory list. Defaults to current workin directory.
        memory_list_filename (str, optional): filename of the memory list. Defaults to 'nucleo.mem'.
        database_location (str, optional): location of the database. Defaults to 'data/memory_template/'.

    Returns:
        dict, 2 x N: lengths of center templates bonds, 1st layer is type, 2nd layer is relative res_id
        dict, 2 x 2 x N1 x N2: lengths of neighbor templates bonds between the pair of (type, res_id) atoms
    """
    if memory_list_dir == 'DEFAULT':
        memory_list_dir = os.getcwd()
    # read the associative memory list files
    memory_list_location = os.path.join(memory_list_dir, memory_list_filename)
    memory_list = pd.read_csv(memory_list_location, skiprows=4, sep='\s+', names=['filename', 'nucleo_center', 'type1', 'res1', 'type2', 'res2_list'])
    
    # bonds dict initialization
    centerBonds = {}
    for dna_type in _DNA_TYPES:
        centerBonds[dna_type] = {}
    neighborBonds = {}
    for dna_type1 in _DNA_TYPES:
        neighborBonds[dna_type1] = {}
        for dna_type2 in _DNA_TYPES:
            neighborBonds[dna_type1][dna_type2] = defaultdict(dict)

    # read each associative memory fragment
    old_filename = None
    for index, row in memory_list.iterrows():
        if old_filename != row['filename']:
            fragment_filename = pkg_resources.resource_filename(__name__, database_location + row['filename'])
            # read the memory file
            frag = Universe(PDBxFile(fragment_filename))
        old_filename = row['filename']

        nucleo_zero = row['nucleo_center']
        type1 = row['type1']
        res1 = row['res1']
        type2 = row['type2']
        res2_list = [int(x) for x in row['res2_list'][1:-1].split(",")]

        

        atom1 = frag.select_atoms(f"name {type1} and resid {res1 + 1}")[0] #openmm index starts from 0, mda from 1
        pos1 = atom1.position / 10 # MDanalysis default units Angstrom, openmm default units nm
        for res2 in res2_list:
            atom2 = frag.select_atoms(f"name {type2} and resid {res2 + 1}")[0] #openmm index starts from 0, mda from 1
            pos2 = atom2.position / 10 # MDanalysis default units Angstrom, openmm default units nm

            rm = np.linalg.norm(pos1 - pos2)
            if type1 == _PRO_TYPE:
                centerBonds[type2][res2 - nucleo_zero] = rm
            else:
                neighborBonds[type1][type2][res1 - nucleo_zero][res2 - nucleo_zero] = rm

    return centerBonds, neighborBonds



import pandas as pd
from .wechrom import *
from .prepare_memory import generate_nucleo_memory_files, get_nucleo_memory_template_bonds
from.utils import _PRO_RES, _PRO_TYPE, _VIRTUAL_TYPE

_NUCLEO_XML = pkg_resources.resource_filename(__name__, 'data/nucleo.xml')

class SingleNucleoSystem(WechromSystem):
    """Build a single-nucleosome WEChroM system with topology and system information for openmm simulations

    Args:
        WechromSystem (object): DNA only system

    Attributes:
        proTypes (list of str): names of protein atoms
        protein (list of int): index of protein atoms
        proRes (list of int): index of protein residues
        pro_Atom2Res (dict): atom id to res id relation for proteins
        pro_Res2Atom (dict): res id to atom id relation for proteins
        nucZero (int): res id of the zero point DNA bp of the nucleosome
    """
    def __init__(self, pdbxFile, xmlFile='DEFAULT', memoryRange=4, nucZero = -1, verbose=False):
        if xmlFile == 'DEFAULT':
            xmlFile = _NUCLEO_XML

        # inherent the parent initialization
        super().__init__(pdbxFile, xmlFile, memoryRange, verbose)

        # atom index for protein
        self.proTypes = [_PRO_TYPE]
        self.protein = [atom.index for atom in self.atoms if atom.name == _PRO_TYPE]
        assert len(self.protein) == 1, "more than one histone core detected"

        # residues for protein
        self.proRes = [residue.index for residue in self.residues if residue.name == _PRO_RES]

        self.pro_Atom2Res = {atom:self.atoms[atom].residue.index for atom in self.protein}
        self.pro_Res2Atom = {self.pro_Atom2Res[atom]:atom for atom in self.pro_Atom2Res}

        # if zero point of the nucleosome DNA is not specified, use the central bp of the DNA molecule
        if nucZero == -1:
            self.nucZero = self.nBp//2
        else:
            self.nucZero = nucZero

        self.virtualSites = [atom.index for atom in self.atoms if atom.name == _VIRTUAL_TYPE]
        # sanity check
        for virtualSite in self.virtualSites:
            assert self.system.isVirtualSite(virtualSite), f"Virtual Sites {virtualSite} not aligned correctly, please check the index in your pdbx file"


    def addExcludVolumeForce(self, kExcl=5856.0, rExclDna=2.07, rExclPro = 26.0, forceName = 'Ex-vol'):
        """Add the connectivity term to the system.
        Overide the original method in WechromSystem"""
        if self.verbose:
            print("Building excluded volume force......", end = ' ')
        
        # set up the excluded volume force
        excl = CustomNonbondedForce("epsilon*step(sigma-r)*(r-sigma)^2; sigma=0.5*(sigma1+sigma2); epsilon=sqrt(epsilon1*epsilon2)")
        excl.addPerParticleParameter("sigma")
        excl.addPerParticleParameter("epsilon")

        # add all DNA atoms
        # note in naked DNA system, all atoms are DNA
        for atom in self.atoms:
            if atom.name in self.dnaTypes:
                excl.addParticle([rExclDna/2, np.sqrt(kExcl)])
            elif atom.name in self.proTypes:
                excl.addParticle([rExclPro/2, np.sqrt(kExcl)])
            elif atom.name == _VIRTUAL_TYPE:
                excl.addParticle([0.0, 0.0])
            else:
                raise Exception(f"Atom name {atom.name} not recognized")

        # exclude nearest neighbors on the same strand
        # those connected by the connectivity term    
        for dnaType in self.dnaTypes:
            for i in range(self.nBp - 1):
                excl.addExclusion(self.dnaStrands[dnaType][i], self.dnaStrands[dnaType][i+1])    
        
        # Set cutoff distance and method
        excl.setCutoffDistance((rExclDna + rExclPro)/2)
        excl.setNonbondedMethod(excl.CutoffNonPeriodic)

        # Add to the system
        self.addForce(excl, forceName)
        if self.verbose:
            print("done")

    def addNucleoCenterMemoryForce(self, centerBonds, centerDepth = 0.4, centerWidth = 3.0, forceName = "cen_nuc"):
        """Add the nucleosome center associative memory term to the system.
        """
        if self.verbose:
            print("Building nucleosome center associative memory force......", end = ' ')
        
        # initiliaze center and neighbor forces
        center = CustomBondForce('-depth_center*exp((r-r0)^2/(-2.0*sigma_center^2))')
        center.addGlobalParameter('depth_center', centerDepth)
        center.addGlobalParameter('sigma_center', centerWidth)
        center.addPerBondParameter('r0')

        proteinAtom = self.protein[0]
        for dnaType in centerBonds:
            for dnaRes in centerBonds[dnaType]:
                targetRes = dnaRes + self.nucZero
                dnaAtom = self.dna_Res2Atom[dnaType][targetRes]
                center.addBond(proteinAtom, dnaAtom, [centerBonds[dnaType][dnaRes]])

        self.addForce(center, forceName)
        if self.verbose:
            print("done")

    def addNucleoNeighborMemoryForce(self, neighborBonds, neighborDepth = 0.2, neighborWidth = 1.0, forceName = "ngb_nuc"):
        """Add the nucleosome neighbor associative memory term to the system.
        """
        if self.verbose:
            print("Building nucleosome neighbor associative memory force......", end = ' ')
        neighbor = CustomBondForce('-depth_neighbor*exp((r-r0)^2/(-2.0*sigma_neighbor^2))')
        neighbor.addGlobalParameter('depth_neighbor', neighborDepth)
        neighbor.addGlobalParameter('sigma_neighbor', neighborWidth)
        neighbor.addPerBondParameter('r0')

        for type1 in neighborBonds:
            for type2 in neighborBonds[type1]:
                for res1 in neighborBonds[type1][type2]:
                    for res2 in neighborBonds[type1][type2][res1]:
                        targetRes1 = res1 + self.nucZero
                        targetRes2 = res2 + self.nucZero
                        dnaAtom1 = self.dna_Res2Atom[type1][targetRes1]
                        dnaAtom2 = self.dna_Res2Atom[type2][targetRes2]
                        neighbor.addBond(dnaAtom1, dnaAtom2, [neighborBonds[type1][type2][res1][res2]])
        self.addForce(neighbor, forceName)
        if self.verbose:
            print("done")
        
    def addDefaultForces(self):
        """Add default forces designed for the wechrom system, including all DNA forces, updated excluded-volume forces and nucleosome associative memory forces
        """
        super().addDefaultForces()

        generate_nucleo_memory_files()
        centerBonds, neighborBonds = get_nucleo_memory_template_bonds()
        self.addNucleoCenterMemoryForce(centerBonds)
        self.addNucleoNeighborMemoryForce(neighborBonds)



from email import header
import pkg_resources
import os
import numpy as np
from collections import defaultdict
from tqdm import tqdm
from .prepare_memory import generate_naked_memory_files, get_naked_memory_template_bonds
from .utils import num_intra_memory_bonds, num_inter_memory_bonds
from .utils import _DNA_S1_TYPE, _DNA_S2_TYPE, _DNA_RES

try:
    # for openmm version >= 7.6
    from openmm import LangevinIntegrator, HarmonicBondForce, CustomNonbondedForce, CustomBondForce, Platform
    from openmm.app import CheckpointReporter, Simulation, DCDReporter, PDBxFile, ForceField, DCDFile, PDBFile
    from openmm.unit import *
except:
    # for openmm version <= 7.5
    from simtk.openmm import LangevinIntegrator, HarmonicBondForce, CustomNonbondedForce, CustomBondForce, Platform
    from simtk.openmm.app import Simulation, DCDReporter, CheckpointReporter, DCDFile, PDBxFile, ForceField, PDBFile
    from simtk.unit import *


_DNA_XML = pkg_resources.resource_filename(__name__, 'data/DNA.xml')

class WechromSystem:
    """ Build a WEChroM system with topology and system information for openmm simulations
    The vanilla version should only have DNA beads
    
    Attributes:
        positions (np.array, N x 3):
            positions of all atoms in the initial pdbx file
        atoms (list of openmm Atoms):
            index of all atoms (DNA)
        residues (list of openmm Residues):
            index of all residues (base pairs)
        dnaStrands (dict of list of int):
            index of all atoms in the two strands I and J
        dnaRes (list of int):
            index of all DNA residues
        nBp (int):
            number of DNA residues
        dna_Atom2Res, dna_Res2Atom (dict of dict):
            atom2res and res2atom relation for the two dna strands
        dnaTypes (list):
            name of the two strands I and J
        pdbx (openmm PDBxFile):
        forcefield (openmm forcefield):
            built on the xml file
        system (openmm system):
            built on topology from pdbx and forcefield

    """
    def __init__(self, pdbxFile, xmlFile='DEFAULT',  memoryRange = 4, verbose = False):
        """Constructs all the necessary attributes to a WEChroM system

        Args:
            pdbxFile (str or file): the pdbx file containing system topology and initial positions
            xmlFile (str or file, optional): the xml file containing openmm forcefield. Defaults to data/DNA.xml.
            verbose(bool, optional): whether to print the information during the simulation. Defaults to True
        """
        # Construct a Topology and atom positions from the pdbx file
        
        # allow pdb filename beyond pdbx/mmcif file
        if isinstance(pdbxFile, str) and pdbxFile.endswith("pdb"):
            self.pdbx = PDBFile(pdbxFile)
        else:
            self.pdbx = PDBxFile(pdbxFile)
        self.topology = self.pdbx.topology
        self.positions = self.pdbx.getPositions(asNumpy=True)
        self.atoms = list(self.topology.atoms())
        self.residues = list(self.topology.residues())
        self.memoryRange = memoryRange

        # Read the xml file to construct a system
        if xmlFile == 'DEFAULT':
            xmlFile = _DNA_XML
        self.forcefield = ForceField(xmlFile)
        self.system = self.forcefield.createSystem(self.topology)
        self.force_Group2Name = {}

        # Construct indexing for DNA particles
        self.constructDNAIndex()

        self.verbose = verbose

    def constructDNAIndex(self):
        """Construct DNA atom and residue index, as well as their relations
        """
        self.dnaTypes = [_DNA_S1_TYPE, _DNA_S2_TYPE]

        # Atom index for each DNA strand
        self.dnaStrands = {}
        for dnaType in self.dnaTypes:
            self.dnaStrands[dnaType] = [atom.index for atom in self.atoms if atom.name == dnaType]
        
        # DNA residues (bairpairs)
        self.dnaRes = [residue.index for residue in self.residues if residue.name == _DNA_RES]
        self.nBp = len(self.dnaRes)

        # Res2Atom and Atom2Res relation for each type
        self.dna_Atom2Res = {}
        self.dna_Res2Atom = {}
        for dnaType in self.dnaTypes:
            self.dna_Atom2Res[dnaType] = {atom:self.atoms[atom].residue.index for atom in self.dnaStrands[dnaType]}
            self.dna_Res2Atom[dnaType] = {self.dna_Atom2Res[dnaType][atom]:atom for atom in self.dna_Atom2Res[dnaType]}
    
    def addForce(self, force, forceName):
        """Add a force to the system, set the force group and store the corresponding str name

        Args:
            forces (openmm force)
            forceName (str)
        """
        nForce = self.system.getNumForces()
        forceGroup = nForce
        force.setForceGroup(forceGroup)

        self.force_Group2Name[forceGroup] = forceName
        self.system.addForce(force)

    def addConnetivityForce(self, kCon=3000.0, bondLength=2.0, forceName = 'Con'):
        """Add a connectivity force to the system

        Args:
            kCon (float, optional): k in harmonic spring 0.5k(r-r0)^2. Defaults to 3000.0.
            bondLength (float, optional): r0 in harmonic spring 0.5k(r-r0)^2. Defaults to 2.0.
            forceName (str, optional): Defaults to 'Connectivity'.
        """
        if self.verbose:
            print("Building connectivity terms......", end = ' ')

        # set up the connectivity force between nearest neighbors on the same strand
        con = HarmonicBondForce()
        for dnaType in self.dnaTypes:
            for i in range(self.nBp - 1):
                con.addBond(self.dnaStrands[dnaType][i], self.dnaStrands[dnaType][i+1], bondLength, kCon)

        # add the connectivity force to our system
        self.addForce(con, forceName)
        if self.verbose:
            print("done")

    def addExcludVolumeForce(self, kExcl=5856.0, rExcl=2.07, forceName = 'Ex-vol'):
        """Add the connectivity term to the system"""
        if self.verbose:
            print("Building excluded volume force......", end = ' ')
        
        # set up the excluded volume force
        excl = CustomNonbondedForce("epsilon*step(sigma-r)*(r-sigma)^2; sigma=0.5*(sigma1+sigma2); epsilon=sqrt(epsilon1*epsilon2)")
        excl.addPerParticleParameter("sigma")
        excl.addPerParticleParameter("epsilon")

        # add all DNA atoms
        # note in naked DNA system, all atoms are DNA
        for atom in self.atoms:
            if atom.name in self.dnaTypes:
                excl.addParticle([rExcl/2, np.sqrt(kExcl)])
            else:
                raise Exception("None DNA atom detected in vanilla WeChroM system")

        # exclude nearest neighbors on the same strand
        # those connected by the connectivity term    
        for dnaType in self.dnaTypes:
            for i in range(self.nBp - 1):
                excl.addExclusion(self.dnaStrands[dnaType][i], self.dnaStrands[dnaType][i+1])    
        
        # Set cutoff distance and method
        excl.setCutoffDistance(rExcl)
        excl.setNonbondedMethod(excl.CutoffNonPeriodic)

        # Add to the system
        self.addForce(excl, forceName)
        if self.verbose:
            print("done")

    def addIntraStrandMemoryForce(self, temPos, depth = 0.3, width= 0.2, forceName = 'Intra_m'):
        """Add the intra-strand associative memory force to the system

        Args:
            temPos (dict): dict, 10 x 2 x memory_range x 3: positions of templates
            depth (float, optional): depth of the Gaussian Well. Defaults to 0.3. unit = kT
            width (float, optional): width of the Gaussian Well. Defaults to 0.2. unit = reduced length
            Available (depth, width) pairs: (0.15, 0.1),(0.2, 0.15), (0.25, 0.17), (0.3, 0.2)
        """
        # Apply the memory template repetitively along the DNA
        if self.verbose:
            print("Building DNA intra-strand associative memory force......", end = ' ')
        # Init the force in Gaussian Well
        intra = CustomBondForce(f'-depth_intra*exp((r-r0)^2/(-2.0*sigma^2))')
        intra.addGlobalParameter('depth_intra', depth)
        intra.addPerBondParameter('r0')
        intra.addPerBondParameter('sigma')

        # Initialize template distances temDist
        # temDist[dnaType][resDiff] should have size nTemplate * (memoryRange - resDiff)
        # in this way, it's equivalent to slide the memory along the DNA
        temDists = {}        
        for dnaType in self.dnaTypes:
            temDists[dnaType] = defaultdict(list)
        # fill in temDist
        for temId in temPos:
            for dnaType in temPos[temId]:
                positions = temPos[temId][dnaType]
                for i, pos1 in enumerate(positions):
                    for j in range(i + 1, len(positions)):
                        temDists[dnaType][j-i].append(np.linalg.norm(pos1 - positions[j]))
        
        # build associative memory bonds from templates
        for dnaType in self.dnaTypes:
            for i in range(self.nBp - 1):
                for j in range(i + 1, min(self.nBp, i + self.memoryRange)):
                    for temDist in temDists[dnaType][j-i]:
                        r0 = temDist
                        sigma = width * (j - i)**0.5
                        intra.addBond(self.dnaStrands[dnaType][i], self.dnaStrands[dnaType][j], [r0, sigma])
        
        # Sanity Check: number of bonds
        nBonds = intra.getNumBonds()
        nBondsExpect = num_intra_memory_bonds(self.nBp, self.memoryRange, len(temPos))
        if nBondsExpect != nBonds:
            print(f"\nALERT! Expect {nBondsExpect} bonds but got {nBonds} bonds for {forceName}")

        # Add to the system
        self.addForce(intra, forceName)

        if self.verbose:
            print("done")

    def addInterStrandMemoryForce(self, temPos, depth = 0.3, width= 0.2, forceName = 'Inter_m'):
        """Add the inter-strand associative memory force to the system

        Args:
            temPos (dict): dict, 10 x 2 x memory_range x 3: positions of templates
            depth (float, optional): depth of the Gaussian Well. Defaults to 0.3. unit = kT
            width (float, optional): width of the Gaussian Well, has a factor of 2 compared with the intra-strand width. Defaults to 0.2. unit = reduced length
            Available (depth, width) pairs: (0.15, 0.1),(0.2, 0.15), (0.25, 0.17), (0.3, 0.2) should keep the same as intra-strand
        """
        # Apply the memory template repetitively along the DNA
        if self.verbose:
            print("Building DNA inter-strand associative memory force......", end = ' ')
        # Init the force in Gaussian Well
        inter = CustomBondForce(f'-depth_inter*exp((r-r0)^2/(-2.0*sigma^2))')
        inter.addGlobalParameter('depth_inter', depth)
        inter.addPerBondParameter('r0')
        inter.addPerBondParameter('sigma')
        
        # inter-strand width factor
        width *= 2

        # Initialize template distances temDist
        # temDist[dnaType][resDiff] should have size nTemplate * (memoryRange - resDiff)
        # in this way, it's equivalent to slide the memory along the DNA
        temDists = {}        
        for dnaType in self.dnaTypes:
            temDists[dnaType] = defaultdict(list)

        # fill in temDist
        I, J = self.dnaTypes
        for temId in temPos:
            iPos = temPos[temId][I]
            jPos = temPos[temId][J]
            for id1 in range(len(iPos)):
                for id2 in range(id1, len(iPos)):
                    temDists[I][id2 - id1].append(np.linalg.norm(iPos[id1] - jPos[id2]))
                    temDists[J][id2 - id1].append(np.linalg.norm(jPos[id1] - iPos[id2]))
        
        # build associative memory bonds from templates
        for i in range(self.nBp - 1):
            for j in range(i + 1, min(self.nBp, i + self.memoryRange)):
                for temDist in temDists[I][j-i]:
                    r0 = temDist
                    sigma = width * (j - i)**0.5
                    inter.addBond(self.dnaStrands[I][i], self.dnaStrands[J][j], [r0, sigma])
                for temDist in temDists[J][j-i]:
                    r0 = temDist
                    sigma = width * (j - i)**0.5
                    inter.addBond(self.dnaStrands[J][i], self.dnaStrands[I][j], [r0, sigma])

        # fill in the same base-pair bonds
        for i in range(self.nBp):
            for temDist in temDists[I][0]:
                r0 = temDist
                sigma = width 
                inter.addBond(self.dnaStrands[I][i], self.dnaStrands[J][i], [r0, sigma])
        
        # Sanity Check: number of bonds
        nBonds = inter.getNumBonds()
        nBondsExpect = num_inter_memory_bonds(self.nBp, self.memoryRange, len(temPos))
        if nBondsExpect != nBonds:
            print(f"\nALERT! Expect {nBondsExpect} bonds but got {nBonds} bonds for {forceName}")

        # Add to the system
        self.addForce(inter, forceName)
        
        if self.verbose:
            print("done")

    def addDefaultForces(self):
        """Add default forces designed for the wechrom system, including the connectivity term, the excluded-volume term, generating associative memory templates and adding inter- and intra- strand associative memory terms
        """
        self.addConnetivityForce()
        self.addExcludVolumeForce()
        generate_naked_memory_files()
        self.temPos = get_naked_memory_template_bonds()
        self.addIntraStrandMemoryForce(self.temPos)
        self.addInterStrandMemoryForce(self.temPos)
            
    def initializeSimulation(self, platform = 'CPU', temperature = 300, collisionRate = 1.0, timeStep = 20.0):
        """Initialize an openmm simulation. 

        Args:
            platform (str, optional): platform. Defaults to 'CPU'. Can be 'CPU', 'CUDA' or 'OpenCL'
            temperature (int, optional): temperature in kelvin. Defaults to 300.
            collisionRate (float, optional): collision rate for the langevin integrator in 1/picosecond. Defaults to 1.0.
            timeStep (float, optional): simulation time step in femtosecond. Defaults to 20.0.

        Attributes:
            platform (openmm platform): CPU, CUDA or OpenCL. Defaults to CPU
            integrator (openmm integrator): a LangevinIntegrator with given temperature, collision rate and timestep
            simulation (openmm simulation): a simulation with wechrom topology, system, instance's integrator and given platform. The initial positions are set by wechrom pdbx file and the initial velocities are set to temperature. Local energy is minimized.
        """


        # set up integrator 
        self.platform = Platform.getPlatformByName(platform)
        self.temperature = temperature / 300 * 120.27  * kelvin # reduced temperature
        self.collisionRate = collisionRate * 2.75 / picosecond # reduced versed time
        self.timestep = timeStep * 0.001 / 2.75  * picoseconds # reduced time
        self.integrator = LangevinIntegrator(self.temperature, self.collisionRate , self.timestep)

        # set up simulation, initial positions and velocities
        self.simulation = Simulation(self.topology, self.system, self.integrator, self.platform)
        self.simulation.context.setPositions(self.positions)
        self.simulation.context.setVelocitiesToTemperature(self.temperature)
        self.simulation.minimizeEnergy() 
        if self.verbose:
            print("Langevin integrator and simulation initialized")

    def runSteps(self, steps = 1000, reportFreq = 100, append = False, outputDir = 'DEFAULT', energyFilename = 'energy.txt', dcdFilename = 'movie.dcd', chkFilename ='checkpt.chk' ):
        """Run simulation on openmm. Should be called after initializeSimulation(). Report energy for each force and trajectories in dcd format at the given report frequency. Save the last checkpoint

        Args:        
            outputDir (str, optional): output directory. Defaults to 'DEFAULT'.
            steps (int, optional): total steps for the simulation. Defaults to 1000.
            reportFreq (int, optional): report frequency. Defaults to 100.
            append (bool, optional): append to an old trajectory and energy or not. Defaults to False.
            energyFilename (str, optional): file name to report energy. Defaults to 'energy.txt'.
            dcdFilename (str, optional): file name to report trajectory. Defaults to 'movie.dcd'.
            chkFilename (str, optional): file name for the last checkpoint. Defaults to 'checkpt.chk'.
        """
        
        if outputDir == 'DEFAULT':
            outputDir = os.getcwd()
        self.outputDir = outputDir

        # convert steps and reportFreq to int
        steps = int(steps)
        reportFreq = int(reportFreq)
        assert steps % reportFreq == 0, "Steps must be devisible by reportFreq"

        if self.verbose:
            print(f"Simulation will take {steps} steps and get reported every {reportFreq} steps")
        # Set up reporters
        # a DCD reporter to store the trajectory 
        # dcd_file = os.path.join(self.outputDir, dcdFilename)
        # self.simulation.reporters.append(DCDReporter(dcd_file, reportFreq, append = append))

        # a checkpoint reporter to store the last checkpoint
        chk_file = os.path.join(self.outputDir, chkFilename)
        self.simulation.reporters.append(CheckpointReporter(chk_file, reportFreq))

        # initialize the energy reporter
        if not append:
            with open(energyFilename, 'w') as fe:
                header ='Steps ' + ' '.join(['{0:<8s}'.format(i) for i in self.force_Group2Name.values()]) + ' Total Kinetics(kT)\n'
                fe.write(header)
        
        if self.verbose:
            print("----------------Simulation Starts----------------")

        # add a dcd writer manually to get it properly closed at the end of simulation
        dcdFilepath = os.path.join(self.outputDir, dcdFilename)
        dcdFilehandle = open(dcdFilepath, 'bw')
        dcdWriter = DCDFile(dcdFilehandle, self.topology, self.timestep)

        # run the simulation
        nSavedFrames = steps // reportFreq
        for i in tqdm(range(nSavedFrames)):
            self.simulation.step(reportFreq)

            # potential energy from each group
            groupEnergy = []
            for group in self.force_Group2Name:
                state = self.simulation.context.getState(getEnergy=True, groups={group})
                groupEnergy.append(state.getPotentialEnergy().value_in_unit(kilojoule_per_mole))
            # total potential energy
            state = self.simulation.context.getState(getEnergy=True, getPositions=True)
            groupEnergy.append(state.getPotentialEnergy().value_in_unit(kilojoule_per_mole))
            # kinetic energy
            groupEnergy.append(state.getKineticEnergy().value_in_unit(kilojoule_per_mole))
            # positions
            positions = state.getPositions(True)
            dcdWriter.writeModel(positions)
            
            with open(energyFilename, 'a') as fe:
                curStep = (i+1)*reportFreq
                line = f'{curStep:<8}' + ' '.join(['{0:<8.2f}'.format(_) for _ in groupEnergy]) + '\n'
                fe.write(line)
        if self.verbose:
            print("\nSimulation done.")
            print(f"Please check your trajectory file {dcdFilename}, energy file {energyFilename}", end = ' ')
            print(f"at your output directory {self.outputDir}")
        # close the dcd file 
        dcdFilehandle.close()
